{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Lists\n",
    "\n",
    "A list is a collection of items where each item holds a relative position with respect to the others. More specifically, we will refer to this type of list as an **unordered list**. We can consider the list as having a first item, a second item, a third item, and so on. We can also refer to the beginning of the list (the first item or the end of the list (the last item).\n",
    "\n",
    "```{mermaid}\n",
    "flowchart LR;\n",
    "    head[/head\\]\n",
    "    a[54]\n",
    "    b[23]\n",
    "    c[12]\n",
    "    d[78]\n",
    "    e[89]\n",
    "    f[42]\n",
    "    null((null))\n",
    "\n",
    "    head-->a\n",
    "    a --next--> b\n",
    "    b --next--> c\n",
    "    c --next--> d\n",
    "    d --next--> e\n",
    "    e --next--> f\n",
    "    f --next--> null\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Unordered List Abstract Data Type\n",
    "\n",
    "The structure of an unordered list, as described above, is a collection of items where each item holds a relative position with respect to the others. Some possible unordered list operations are given below.\n",
    "\n",
    "- `UnorderedList()` creates a new unordered list that is empty. It needs no parameters and returns an empty list.\n",
    "- `add(item)` adds a new item to the list. It needs the item and returns nothing. Assume the item is not already in the list.\n",
    "- `remove(item)` removes the item from the list. It needs the item and modifies the list. Assume the item is present in the list.\n",
    "- `search(item)` searches for the item in the list. It needs the item and returns a boolean value.\n",
    "- `is_empty()` tests to see whether the list is empty. It needs no parameters and returns a boolean value.\n",
    "- `size()` returns the number of items in the list. It needs no parameters and returns an integer.\n",
    "- `append(item)` adds a new item to the end of the list making it the last item in the collection. It needs the item and returns nothing. Assume the item is not already in the list.\n",
    "- `index(item)` returns the position of item in the list. It needs the item and returns the index. Assume the item is in the list.\n",
    "- `insert(pos,item)` adds a new item to the list at position pos. It needs the item and returns nothing. Assume the item is not already in the list and there are enough existing items to have position pos.\n",
    "- `pop()` removes and returns the last item in the list. It needs nothing and returns an item. Assume the list has at least one item.\n",
    "- `pop(pos)` removes and returns the item at position pos. It needs the position and returns the item. Assume the item is in the list.\n",
    "\n",
    "### Implementing an Unordered List\n",
    "\n",
    "In order to implement an unordered list, we will construct what is commonly known as a linked list. Recall that we need to be sure that we can maintain the relative positioning of the items. However, there is no requirement that we maintain that positioning in contiguous memory. If we can maintain some explicit information in each item, namely the location of the next item, then the relative position of each item can be expressed by simply following the link from one item to the next.\n",
    "\n",
    "However location of the first item of the list must be explicitly specified. Once we know where the first item is, the first item can tell us where the second is, and so on. The external reference is often referred to as the **head** of the list. Similarly, the last item needs to know that there is no next item.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {
    "vscode": {
     "languageId": "python"
    }
   },
   "outputs": [],
   "source": [
    "class Node:\n",
    "\n",
    "    def __init__(self, data):\n",
    "        self.data = data\n",
    "        self.next_node = None\n",
    "    \n",
    "    def get_data(self):\n",
    "        return self.data\n",
    "    \n",
    "    def get_next_node(self):\n",
    "        return self.next_node\n",
    "\n",
    "    def set_data(self, data):\n",
    "        self.data = data\n",
    "    \n",
    "    def set_next_node(self, next_node):\n",
    "        self.next_node = next_node"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "vscode": {
     "languageId": "python"
    }
   },
   "outputs": [],
   "source": [
    "# Todo: Fix methods to have O(1) append\n",
    "class UnorderedList:\n",
    "\n",
    "    def __init__(self):\n",
    "        self.num_nodes = 0\n",
    "        self.head = None\n",
    "        self.tail = self.head\n",
    "\n",
    "    def is_empty(self):\n",
    "        return self.head == None\n",
    "    \n",
    "    def add(self, data):\n",
    "        self.num_nodes += 1\n",
    "\n",
    "        n = Node(data)\n",
    "        \n",
    "        if not self.tail:\n",
    "            self.tail = n\n",
    "        \n",
    "        n.set_next_node(self.head)\n",
    "        self.head = n\n",
    "    \n",
    "    def remove(self, data):\n",
    "        self.num_nodes -= 1\n",
    "\n",
    "        previous = None\n",
    "        current = self.head\n",
    "\n",
    "        while current:\n",
    "            if current.get_data() == data:\n",
    "                if not previous:\n",
    "                    self.head = current.get_next_node()\n",
    "                else:\n",
    "                    previous.set_next_node(current.get_next_node())\n",
    "\n",
    "            previous = current\n",
    "            current = current.get_next_node()\n",
    "\n",
    "    def search(self, data):\n",
    "        current = self.head\n",
    "\n",
    "        while current:\n",
    "            if current.get_data() == data:\n",
    "                return True\n",
    "            current = current.get_next_node()\n",
    "        \n",
    "        return False\n",
    "    \n",
    "    def size(self):\n",
    "        return self.num_nodes\n",
    "\n",
    "    \n",
    "    def append(self, data):\n",
    "        self.num_nodes += 1\n",
    "\n",
    "        n = Node(data)\n",
    "\n",
    "        if self.tail:\n",
    "            self.tail.set_next_node(n)\n",
    "        else:\n",
    "            self.head = n\n",
    "    \n",
    "    def index(self, data):\n",
    "        index = -1\n",
    "        current = self.head\n",
    "\n",
    "        while current:\n",
    "            index += 1\n",
    "            if current.get_data() == data:\n",
    "                return index\n",
    "            current = current.get_next_node()\n",
    "        \n",
    "        return index\n",
    "    \n",
    "    def insert(self, pos, data):\n",
    "        self.num_nodes += 1\n",
    "\n",
    "        index = -1\n",
    "        n = Node(data)\n",
    "        current = self.head\n",
    "\n",
    "        while current:\n",
    "            index += 1\n",
    "            if index == pos:\n",
    "                n.set_next_node(current.get_next_node())\n",
    "                current.set_next_node(n)\n",
    "\n",
    "            current = current.get_next_node()\n",
    "    \n",
    "    def pop(self, pos=-1):\n",
    "        self.num_nodes -= 1\n",
    "\n",
    "        current = self.head\n",
    "        \n",
    "        pos = pos if pos >= 0 else self.num_nodes + pos\n",
    "\n",
    "        for i in range(pos):\n",
    "            current = current.get_next_node()\n",
    "\n",
    "        data = current.get_next_node().get_data()\n",
    "        \n",
    "        current.set_next_node(None)\n",
    "        return data\n",
    "\n",
    "    def __str__(self):\n",
    "        s = \"\"\n",
    "        current = self.head\n",
    "        while current != None:\n",
    "            s += f\"{current.get_data()} -> \"\n",
    "            current = current.get_next_node()\n",
    "\n",
    "        return s.strip(\" -> \")\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Ordered List Abstract Data Type\n",
    "\n",
    "The structure of an ordered list is a collection of items where each item holds a relative position that is based upon some underlying characteristic of the item. The ordering is typically either ascending or descending and we assume that list items have a meaningful comparison operation that is already defined. Many of the ordered list operations are the same as those of the unordered list.\n",
    "\n",
    "- `OrderedList()` creates a new ordered list that is empty. It needs no parameters and returns an empty list.\n",
    "- `add(item)` adds a new item to the list. It needs the item and returns nothing. Assume the item is not already in the list.\n",
    "- `remove(item)` removes the item from the list. It needs the item and modifies the list. Assume the item is present in the list.\n",
    "- `search(item)` searches for the item in the list. It needs the item and returns a boolean value.\n",
    "- `is_empty()` tests to see whether the list is empty. It needs no parameters and returns a boolean value.\n",
    "- `size()` returns the number of items in the list. It needs no parameters and returns an integer.\n",
    "- `index(item)` returns the position of item in the list. It needs the item and returns the index. Assume the item is in the list.\n",
    "- `pop()` removes and returns the last item in the list. It needs nothing and returns an item. Assume the list has at least one item.\n",
    "- `pop(pos)` removes and returns the item at position pos. It needs the position and returns the item. Assume the item is in the list.\n",
    "\n",
    "### Implementing an Ordered List\n",
    "\n",
    "In order to implement the ordered list, we must remember that the relative positions of the items are based on some underlying characteristic. The ordered list of integers given above (17, 26, 31, 54, 77, and 93) can be represented by a linked structure. Again, the node and link structure is ideal for representing the relative positioning of the items.\n",
    "\n",
    "To implement the OrderedList class, we will use the same technique as seen previously with unordered lists. Once again, an empty list will be denoted by a head reference to None."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "vscode": {
     "languageId": "python"
    }
   },
   "outputs": [],
   "source": [
    "# Todo: Fix methods to have O(1) append\n",
    "class OrderedList:\n",
    "\n",
    "    def __init__(self):\n",
    "        self.num_nodes = 0\n",
    "        self.head = None\n",
    "        self.tail = self.head\n",
    "\n",
    "    def is_empty(self):\n",
    "        return self.head == None\n",
    "    \n",
    "    def add(self, data):\n",
    "        self.num_nodes += 1\n",
    "\n",
    "        n = Node(data)\n",
    "        \n",
    "        if not self.tail:\n",
    "            self.tail = n\n",
    "        \n",
    "        n.set_next_node(self.head)\n",
    "        self.head = n\n",
    "    \n",
    "    def remove(self, data):\n",
    "        self.num_nodes -= 1\n",
    "\n",
    "        previous = None\n",
    "        current = self.head\n",
    "\n",
    "        while current:\n",
    "            if current.get_data() == data:\n",
    "                if not previous:\n",
    "                    self.head = current.get_next_node()\n",
    "                else:\n",
    "                    previous.set_next_node(current.get_next_node())\n",
    "\n",
    "            previous = current\n",
    "            current = current.get_next_node()\n",
    "\n",
    "    def search(self, data):\n",
    "        current = self.head\n",
    "\n",
    "        while current:\n",
    "            if current.get_data() == data:\n",
    "                return True\n",
    "            current = current.get_next_node()\n",
    "        \n",
    "        return False\n",
    "    \n",
    "    def size(self):\n",
    "        return self.num_nodes\n",
    "\n",
    "    def index(self, data):\n",
    "        index = -1\n",
    "        current = self.head\n",
    "\n",
    "        while current:\n",
    "            index += 1\n",
    "            if current.get_data() == data:\n",
    "                return index\n",
    "            current = current.get_next_node()\n",
    "        \n",
    "        return index\n",
    "\n",
    "    def __str__(self):\n",
    "        s = \"\"\n",
    "        current = self.head\n",
    "        while current != None:\n",
    "            s += f\"{current.get_data()} -> \"\n",
    "            current = current.get_next_node()\n",
    "\n",
    "        return s.strip(\" -> \")\n"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## DoublyLinked List Abstract Data Type\n",
    "\n",
    "The structure of an DoublyLinked list is a collection of items where each item holds a relative position that is based upon some underlying characteristic of the item. The ordering is typically either ascending or descending and we assume that list items have a meaningful comparison operation that is already defined. Many of the ordered list operations are the same as those of the unordered list.\n",
    "\n",
    "- `DoublyLinkedList()` creates a new DoublyLinked list that is empty. It needs no parameters and returns an empty list.\n",
    "- `add(item)` adds a new item to the list. It needs the item and returns nothing. Assume the item is not already in the list.\n",
    "- `remove(item)` removes the item from the list. It needs the item and modifies the list. Assume the item is present in the list.\n",
    "- `search(item)` searches for the item in the list. It needs the item and returns a boolean value.\n",
    "- `is_empty()` tests to see whether the list is empty. It needs no parameters and returns a boolean value.\n",
    "- `size()` returns the number of items in the list. It needs no parameters and returns an integer.\n",
    "- `index(item)` returns the position of item in the list. It needs the item and returns the index. Assume the item is in the list.\n",
    "- `pop()` removes and returns the last item in the list. It needs nothing and returns an item. Assume the list has at least one item.\n",
    "- `pop(pos)` removes and returns the item at position pos. It needs the position and returns the item. Assume the item is in the list.\n",
    "\n",
    "### Implementing an DoublyLinkedList\n",
    "\n",
    "In order to implement the ordered list, we must remember that the relative positions of the items are based on some underlying characteristic. The ordered list of integers given above (17, 26, 31, 54, 77, and 93) can be represented by a linked structure. Again, the node and link structure is ideal for representing the relative positioning of the items.\n",
    "\n",
    "To implement the OrderedList class, we will use the same technique as seen previously with unordered lists. Once again, an empty list will be denoted by a head reference to None."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "# This is an input class. Do not edit.\n",
    "class Node:\n",
    "    def __init__(self, value):\n",
    "        self.value = value\n",
    "        self.prev = None\n",
    "        self.next = None\n",
    "\n",
    "\n",
    "# Feel free to add new properties and methods to the class.\n",
    "class DoublyLinkedList:\n",
    "    def __init__(self):\n",
    "        self.head = None\n",
    "        self.tail = None\n",
    "\n",
    "    def setHead(self, node):\n",
    "\n",
    "        node.next\n",
    "\n",
    "    def setTail(self, node):\n",
    "        # Write your code here.\n",
    "        pass\n",
    "\n",
    "    def insertBefore(self, node, nodeToInsert):\n",
    "        # Write your code here.\n",
    "        pass\n",
    "\n",
    "    def insertAfter(self, node, nodeToInsert):\n",
    "        # Write your code here.\n",
    "        pass\n",
    "\n",
    "    def insertAtPosition(self, position, nodeToInsert):\n",
    "        # Write your code here.\n",
    "        pass\n",
    "\n",
    "    def removeNodesWithValue(self, value):\n",
    "        start = self.head\n",
    "        while start:\n",
    "            if start.value == value:\n",
    "                break\n",
    "            start = start.next\n",
    "\n",
    "        next = start.next\n",
    "        previous = start.prev\n",
    "\n",
    "        start.next = start.prev = None\n",
    "\n",
    "        previous.next = next\n",
    "        next.prev = previous\n",
    "\n",
    "    def remove(self, node):\n",
    "        start = self.head\n",
    "        while start:\n",
    "            if start == node:\n",
    "                break\n",
    "            start = start.next\n",
    "\n",
    "        next = start.next\n",
    "        previous = start.prev\n",
    "\n",
    "        start.next = start.prev = None\n",
    "\n",
    "        previous.next = next\n",
    "        next.prev = previous\n",
    "        \n",
    "\n",
    "    def containsNodeWithValue(self, value):\n",
    "        node = self.head\n",
    "        while node:\n",
    "            if node.value == value:\n",
    "                return True\n",
    "            node = node.next\n",
    "        return False\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.4"
  },
  "orig_nbformat": 2,
  "vscode": {
   "interpreter": {
    "hash": "767d51c1340bd893661ea55ea3124f6de3c7a262a8b4abca0554b478b1e2ff90"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
